//-----------------------------------------------------------------------------
// abon.c
// Adaboost evaluator plugin loader
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: abon.cpp 58 2009-04-02 09:06:59Z ijuranek $
//-----------------------------------------------------------------------------
#include <abon.h>

#include <assert.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

abonErrorCode trivialSetVerbosity(abonDetector *detector, unsigned verbosity)
{
	detector->verbosity = verbosity;
	return aboneOk;
}

abonErrorCode trivialSetTimer(abonDetector *detector, Timer *timer)
{
	detector->timer = timer;
	return aboneOk;
}

abonErrorCode trivialSetClassifier(abonDetector *detector, TClassifier *classifier)
{
	detector->classifier = classifier;
	return aboneOk;
}

abonErrorCode trivialSetStep(abonDetector *detector, unsigned hstep, unsigned vstep)
{
	detector->hstep = hstep;
	detector->vstep = vstep;
	return aboneOk;
}

abonErrorCode trivialSetSize(abonDetector *detector, unsigned width, unsigned height, unsigned pitch)
{
	detector->width = width;
	detector->height = height;
	detector->pitch = pitch;
	return aboneOk;
}

// Emulation of abonProcess by abonStart

abonImageData *processByStartGet(void *param)
{
	abonImageData **imagePtr = (abonImageData**)param;
	abonImageData *image = *imagePtr;
	*imagePtr = NULL;
	return image;
}

abonErrorCode processByStartPut(void *param, abonImageData *image)
{
	return aboneOk;
}

abonErrorCode processByStart(abonDetector *detector, abonImageData *image)
{
	return abonStart(detector, processByStartGet, processByStartPut, (void*)&image);
}

// Emulation of abonStart by abonProcess

abonErrorCode startByProcess(abonDetector *detector, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
	abonImageData *image;
	while((image = get(param)) != NULL)
	{
		abonErrorCode error = abonProcess(detector, image);
		if(error != aboneOk)
		{
			return error;
		}
		error = put(param, image);
		if(error != aboneOk)
		{
			return error;
		}
	}
	return aboneOk;
}

abonErrorCode abonValidateDetector(abonDetector *detector)
{
	abonErrorCode retval = aboneOk;
	if(detector == NULL)
	{
		fputs("abonCreate failed!\n", stderr);
		return aboneError;
	}
	if(detector->setVerbosity == NULL)
	{
		detector->setVerbosity = trivialSetVerbosity;
	}
	if(detector->setTimer == NULL)
	{
		detector->setTimer = trivialSetTimer;
	}
	if(detector->setClassifier == NULL)
	{
		detector->setClassifier = trivialSetClassifier;
	}
	if(detector->setStep == NULL)
	{
		detector->setStep = trivialSetStep;
	}
	if(detector->setSize == NULL)
	{
		detector->setSize = trivialSetSize;
	}
	if((detector->process == NULL) && (detector->start == NULL))
	{
		fputs("both abonStart and abonProcess unitialized!\n", stderr);
		retval = aboneError;
	}
	else
	{
		if(detector->process == NULL)
		{
			detector->process = processByStart;
		}
		if(detector->start == NULL)
		{
			detector->start = startByProcess;
		}
	}
	if(detector->release == NULL)
	{
		fputs("abonRelease unitialized!\n", stderr);
		retval = aboneError;
	}
	return retval;
}

// Enum translation

static const char *const errorCodeStringTable[] = {
	"No error",
	"Invalid parameter",
	"Invalid size",
	"Just error"
};

const char *abonTranslateError(abonErrorCode error)
{
	assert(sizeof(errorCodeStringTable) == sizeof(const char*)*aboneNumErrorCodes);
	return error < aboneNumErrorCodes ? errorCodeStringTable[error] : "Unknown error";
}

static const char *const timeStampStringTable[] = {
	"Total time",
	"Pyramid time",
	"Convolve time",
	"Detect time",
	"Aux time 1",
	"Aux time 2"
};

const char *abonTranslateTimeStamp(abonTimeStamps timestamp)
{
	assert(sizeof(timeStampStringTable) == sizeof(const char*)*abontsNumTimeStamps);
	return timestamp < abontsNumTimeStamps ? timeStampStringTable[timestamp] : NULL;
}

// Plugin loading

#ifdef _WIN32

#include <windows.h>

abonDetector *abonLoad(const char *lib)
{
	char libname[256];
	sprintf(libname, "lib%s.dll", lib);
	HMODULE library = LoadLibrary(libname);
	if(library != NULL)
	{
		abonFunCreate *create = (abonFunCreate*)GetProcAddress(library, "abonCreate");
		if(create != NULL)
		{
			abonDetector *detector = create();
			if(abonValidateDetector(detector) == aboneOk)
			{
				detector->library = (void*)library;
				detector->create = create;
				return detector;
			}
		}
		else
		{
			fputs("GetProcAddress failed!\n", stderr);
		}
		if(FreeLibrary(library) == FALSE)
		{
			fputs("FreeLibrary failed!\n", stderr);
		}
	}
	else
	{
		fputs("LoadLibrary failed!\n", stderr);
	}
	return NULL;
}

void abonUnload(struct abonDetector *detector)
{
	HMODULE library = (HMODULE)detector->library;
	if(detector->release(detector) != aboneOk)
	{
		fputs("abonRelease failed!\n", stderr);
	}
	if((library != NULL) && (FreeLibrary(library) == FALSE))
	{
		fputs("FreeLibrary failed!\n", stderr);
	}
}

#else

#include <dlfcn.h>

abonDetector *abonLoad(const char *lib)
{
	char libname[256];
	sprintf(libname, "lib%s.so", lib);
	void *library = dlopen(libname, RTLD_LAZY | RTLD_LOCAL);
	if(library != NULL)
	{
		abonFunCreate *create = (abonFunCreate*)dlsym(library, "abonCreate");
		if(create != NULL)
		{
			abonDetector *detector = create();
			if(abonValidateDetector(detector) == aboneOk)
			{
				detector->library = (void*)library;
				detector->create = create;
				return detector;
			}
		}
		else
		{
			//fputs("dlsym failed!\n", stderr);
            fprintf(stderr, "'%s' in not ABON module.\n", libname);
		}
		if(dlclose(library) != 0)
		{
			fputs("dlclose failed!\n", stderr);
		}
	}
	else
	{
        fprintf(stderr, "Cannot load '%s'\n", libname);
	}
	return NULL;
}

void abonUnload(struct abonDetector *detector)
{
	void *library = detector->library;
	if(detector->release(detector) != aboneOk)
	{
		fputs("abonRelease failed!\n", stderr);
	}
	if(dlclose(library) != 0)
	{
		fputs("dlclose failed!\n", stderr);
	}
}

#endif

#ifdef __cplusplus
}
#endif
