#include "io.h"

#include <algorithm>
#include <cmath>

#include <cv.h>
#include <cxcore.h>

using namespace std;

template <typename T>
static inline T sqr(const T x)
{
	return x*x;
}

// Distance of centers of two rectangles
static inline float rectDist(CvRect & a, CvRect & b)
{
    return sqrt(static_cast<float>(sqr(a.x - b.x) + sqr(a.y - b.y)));
}

static inline float rectOverlap(CvRect & U, CvRect & V)
{
	float d = rectDist(U, V);

	int ru = (U.width + U.height);
	int rv = (V.width + V.height);

	float r1 = static_cast<float>(min(ru, rv))*0.5f;
	float r2 = static_cast<float>(max(ru, rv))*0.5f;

	return (d > r1+r2) ? 0 : (r1/r2) * (1.0f - (d / (r1+r2)));
}


static inline CvRect detectionRect(abonDetection * detection)
{
	return cvRect(detection->x, detection->y, detection->width, detection->height);
}

struct Bin
{
	//for detection
	unsigned bin;
	float overlap;
	//for bin with same index
	unsigned count;
	unsigned detection;
};

// Clustering and non-maxima supression
static inline unsigned groupDetections(abonDetection *detections, const unsigned count, const float threshold, const unsigned minGroup)
{
	if (count == 0)
		return 0;

	vector<Bin> bins(count);

	for (unsigned i = 0; i < count; ++i)
	{
        Bin bin = {i, 0, 0, i};
        bins[i] = bin;
	}
	
    // Assign a bin to each detection - actual clustering
    for (unsigned i = 0; i < count-1; ++i)
    {
        CvRect U = detectionRect(detections + i);

        for (unsigned j = i+1; j < count; ++j)
        {
            CvRect V = detectionRect(detections + j);

			float overlap = rectOverlap(U, V);

            if (overlap > threshold && overlap > bins[j].overlap)
            {
                bins[j].bin = bins[i].bin;
                bins[j].overlap = overlap;
            }
        }
    }

    // Select detection with max response from each cluster
    for (unsigned i = 0; i < count; ++i)
    {
        unsigned bin = bins[i].bin;
		unsigned det = bins[bin].detection;

		++bins[det].count;

		if (detections[i].response > detections[det].response)
        {
            bins[bin].detection = i;
			//move count to group maximum
			bins[i].count = bins[det].count;
			bins[det].count = 0;
        }
    }

	// Move group maximas to the front
	unsigned ngroup = count;//first nongroup detection
	for (unsigned i = 0; i < count; ++i)
	{
		if (bins[i].count > minGroup)//at i is group
		{
			if (i > ngroup)//some nongroups precede this
			{
				//move group to front
				swap(detections[i], detections[ngroup]);
				++ngroup;//all between ngroup and i are nongroups
			}
		}
		else//at i isn't group
		{
			if (i < ngroup)
			{
				ngroup = i;//set first nongroup
			}
		}
	}

	return ngroup;//first nongroup means group count
}


InputOutput::InputOutput(const Config &cfg, abonDetector *det, Timer *t)
: config(cfg), detector(det), timer(t)
, quit(false)
, input(NULL), output(NULL), frames(0), image(NULL)
{
}

InputOutput::~InputOutput() throw()
{
}

bool InputOutput::runWithIO(Input *in, Output *out)
{
	if(timer != NULL) timerReset(timer);
	memset(times, 0, sizeof(times));
	input = in;
	output = out;
	frames = 0;
	faces = 0;
	if((image = input->getImage(manager, config.start)) != NULL)
	{
		abonSetSize(detector, image->width, image->height, image->pitch);
		abonSetVerbosity(detector, config.verbose);
		abonStart(detector, ioGetImage, ioPutImage, (void*)this);
		//print times
		cout << "Times for " << frames << " frames : \n";
		for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
		{
			cout << '\t' << abonTranslateTimeStamp(static_cast<abonTimeStamps>(i))
				<< " : " << times[i] << "s, " << times[i]/frames << "s/frame\n";
		}
		return true;
	}
	else
	{
		return false;
	}
}

abonImageData *InputOutput::ioGetImage(void *param)
{
	InputOutput *io = (InputOutput *)param;
	Image *retval = NULL;
	if(io != NULL
		&& !io->quit//q pressed
		&& (io->image != NULL)//end of input
		&& ((io->config.frames == 0) || (io->frames < io->config.frames))//frame count
		)
	{
		retval = io->image;
		//read next input image
		io->image = io->input->getImage(io->manager, io->config.step);
		//start total time
		if(io->timer != NULL)
		{
			retval->timestamps[abontsTotal].start = timerTime(io->timer);
		}
		//convert to grayscale
		cvCvtColor(retval->image, retval->grayscale, CV_RGB2GRAY);
		//smooth
		if(io->config.smooth > 0)
		{
			cvSmooth(retval->grayscale, retval->grayscale, CV_GAUSSIAN, 2*io->config.smooth+1);
		}
		//read next frame
		++io->frames;
	}
	return (abonImageData*)retval;
}

void drawDetections(IplImage * inputFrame, abonDetection *detections, unsigned detection_count,
        const CvScalar & color, bool outline, bool printResponse, const CvFont * font)
{
    for (unsigned i = 0; i != detection_count; ++i)
    {
        abonDetection & d = detections[i];
        CvRect r = detectionRect(&d);
        float angle = d.angle;
        float response = d.response;

        r.x += r.width/2;
        r.y += r.height/2;

        // Prepare rotation matrix
        float _negRot1M[6];
        CvMat negRot1M = cvMat(2, 3, CV_32FC1, _negRot1M); // Origin rotation matrix
        cv2DRotationMatrix( cvPoint2D32f(0,0), -angle, 1.0, &negRot1M);

        // Left and Down vectors
        float _right[3] = {float(r.width), 0, 1}, _right1[3];
        float _down[3] = {0, float(r.height), 1}, _down1[3];

        CvMat right = cvMat(3, 1, CV_32F, _right);
        CvMat right1 = cvMat(2, 1, CV_32F, _right1);
        CvMat down = cvMat(3, 1, CV_32F, _down);
        CvMat down1 = cvMat(2, 1, CV_32F, _down1);

        cvMatMul(&negRot1M, &right, &right1);
        cvMatMul(&negRot1M, &down, &down1);

        CvPoint TL = cvPoint(int(r.x - _right1[0]/2 - _down1[0]/2), int(r.y - _right1[1]/2 - _down1[1]/2));
        CvPoint TR = cvPoint(int(r.x + _right1[0]/2 - _down1[0]/2), int(r.y + _right1[1]/2 - _down1[1]/2));
        CvPoint BL = cvPoint(int(r.x - _right1[0]/2 + _down1[0]/2), int(r.y - _right1[1]/2 + _down1[1]/2));
        CvPoint BR = cvPoint(int(r.x + _right1[0]/2 + _down1[0]/2), int(r.y + _right1[1]/2 + _down1[1]/2));

        // Black outline
        if (outline)
        {
            cvLine(inputFrame, TL, TR, cvScalar(0,0,0), 5);
            cvLine(inputFrame, TR, BR, cvScalar(0,0,0), 5);
            cvLine(inputFrame, BR, BL, cvScalar(0,0,0), 5);
            cvLine(inputFrame, BL, TL, cvScalar(0,0,0), 5);
        }

        // line
        cvLine(inputFrame, TL, TR, color);
        cvLine(inputFrame, TR, BR, color);
        cvLine(inputFrame, BR, BL, color);
        cvLine(inputFrame, BL, TL, color);

        if (font && printResponse)
        {
            char responseStr[256];
            sprintf(responseStr, "%3.2f", response);
            CvPoint textPos = TL;
            textPos.y -= 5;
            cvPutText(inputFrame, responseStr, TL, font, color);
        }

    } // Result loop
}

abonErrorCode InputOutput::ioPutImage(void *param, abonImageData *header)
{
	InputOutput *io = (InputOutput*)param;
	Image *image = (Image*)header;
	if((io == NULL) || (image == NULL))
	{
		return aboneParam;
	}
/*	
	if (!image->nmsDone) // NMS was not performed by the detection engine
	{
		image->detectionCount = groupDetections(image->detections, image->detectionCount, io->config.groupThreshold, io->config.groupMinCount);
		io->faces += image->detectionCount;
	}
*/
	if(io->timer != NULL)
	{
		header->timestamps[abontsTotal].end = timerTime(io->timer);
	}

	//draw all detections
	drawDetections(image->image, header->detections, header->detectionCount, cvScalar(255,255,255, 128), false, false, 0);
	//TODO
	//show final image
	if(io->config.show)
	{
		cvShowImage("Output", image->image);
		if(cvWaitKey(10) == 'q')
			io->quit = true;
	}
	//add time
	for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
	{
		io->times[i] += (image->timestamps[i].end - image->timestamps[i].start);
	}
	//output
	if(io->output != NULL)
	{
		io->output->putImage(image);
	}
	//TODO : output
	io->manager.release(image);
	return aboneOk;
}
